/* $Id: which.c,v 1.3 2001/04/17 10:17:54 wulf Exp $ */

/*
 * Copyright (C) 2001 Berndt Josef Wulf <wulf@ping.net.au>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/sysctl.h>

#include <err.h>
#include <fcntl.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>

#define ALLPATH 0x01
#define VERBOSE 0x02
#define SILENT  0x04
#define PASSIVE 0x08
#define MISSED  0x40
#define FOUND   0x80

#define HT      0x09

typedef struct ALIST {
  struct ALIST *next;
  char alias[PATH_MAX];
} ALIST;

char *get_shell(void);
ALIST *get_aliases(char *);
char *find_alias(ALIST *, char *);
void free_list(ALIST *);
void usage(void);

int main(int argc, char **argv)
{
  int  pathlen;
  char flag = 0;
  char *shell;
  char *envpath;
  char *path;
  char *buf;
  char *ptr;
  char filename[PATH_MAX];
  char cmdbuf[PATH_MAX];
  char *alias;
  ALIST *alist;

  /* exit silently when no arguments are given */
  if (argc < 2)
    exit(EXIT_FAILURE); 

  /* evaluate options */
  while (getopt(argc, argv, "ahpsv") != -1)
  {
    switch (optopt)
    {
      case 'a':
        flag |= ALLPATH;
        break;
      case 'v':
        flag |= VERBOSE;
        break;
      case 'p':
        flag |= PASSIVE;
        break;
      case 's':
        flag |= SILENT;
        break;
      case 'h':
	usage();
	exit(EXIT_SUCCESS);
	break;
      default:
        errx(EXIT_FAILURE,"Usage: which [-apsvh] program ...\n");
        break;
    } 
  }

  argc -= optind;
  argv += optind;

  /* ensure we have at least one program argument */
  if (!argc)
    errx(EXIT_FAILURE, "Please provide at least one \"program\" argument");

  /* get PATH environment variable */
  if ((envpath = getenv("PATH")) == 0)
    errx(EXIT_FAILURE, "Environment variable PATH not set");

  pathlen = strlen(envpath) + 1;

  /* get shell of parent process */
  shell = get_shell();

  /* get alias list */
  if (flag & PASSIVE)
    snprintf(cmdbuf, PATH_MAX,"%s -c alias", shell);
  else
    if (strstr(shell,"bash"))
      snprintf(cmdbuf, PATH_MAX,"echo alias | %s -i", shell);
    else
      snprintf(cmdbuf, PATH_MAX,"echo alias | %s", shell);

  alist = get_aliases(cmdbuf);

  /* allocate a buffer */
  if ((buf = malloc(pathlen)) == 0)
  {
    free_list(alist);
    errx(EXIT_FAILURE, "Can't allocate buffer memory");
  }

  /* loop until argument list is exhausted */
  while ( argc > 0 )
  {
    /* Reset FOUND flag */
    flag &= 0x7f;

    /* check for aliases */
    alias = find_alias(alist, *argv);

    if (alias != 0)
    {
      if (!(flag & SILENT))
        printf("alias: %s\n", alias);
      flag |= FOUND;
    }

    path = buf;
    memcpy(path,envpath, pathlen);

    /* search in paths defined by PATH variable */
    while ((ptr = strsep(&path,":")) != 0) 
    {
      /* leading, double and trailing colon set path to current directory */
      if (ptr[0] == NULL)
        snprintf(filename, PATH_MAX, "./%s", *argv);
      else
      	snprintf(filename, PATH_MAX, "%s/%s", ptr, *argv);

      /* check if program exists */
      if (access(filename, X_OK) != -1) 
      {
        flag |= FOUND;

        if (!(flag & SILENT))
          printf("%s\n",filename);

        if (!(flag & ALLPATH))
          break;
      }
    }

    /* set MISSED flag if program wasn't found */
    if (!(flag & FOUND))
    {
      flag |= MISSED;

      if ((flag & VERBOSE) && !(flag & SILENT))
        printf("%s: Command not found\n", *argv);
    }

    argc--;
    argv++;
  }

  free_list(alist);
  free(buf);

  if (flag & MISSED)
    return EXIT_FAILURE;

  return EXIT_SUCCESS;
}

char *
get_shell(void)
{
  int mib[6];
  size_t len;
  struct kinfo_proc2 proc;

  /* get line command of parent process */
  len = sizeof(proc);
  mib[0] = CTL_KERN;
  mib[1] = KERN_PROC2;
  mib[2] = KERN_PROC_PID;
  mib[3] = getppid();
  mib[4] = len;
  mib[5] = 1;

  if (sysctl(mib, 6 , &proc, &len, NULL, 0) < 0)
    errx(EXIT_FAILURE, "Failed to process sysctl request");

  return strdup(proc.p_comm);
}

ALIST *
get_aliases(char cmdbuf[PATH_MAX])
{
  size_t cnt;
  char *buf;
  ALIST *alist;
  ALIST *head;
  FILE *s;

  /* init aliases list */
  if ((alist = malloc(sizeof(ALIST))) == 0)
    errx(EXIT_FAILURE, "Can't allocate memory to initiate alias list");

  alist->next =  NULL;
  head = alist;

  if ((s = popen(cmdbuf,"r")) == 0)
  {
    free_list(alist);
    errx(EXIT_FAILURE, "Can't create new process");
  }

  for (;;)
  {
    buf = fgetln(s, &cnt);

    if (feof(s))
      break;

    snprintf(head->alias, cnt, "%s", buf);
    if ((head->next = malloc(sizeof(ALIST))) == 0)
    {
      free_list(alist);
      errx(EXIT_FAILURE, "Can't allocate memory to add new list element");
    }

    head = head->next;
    head->next = NULL;
  }

  fflush(s);
  pclose(s);
  return alist;
}

char *
find_alias(ALIST *alist, char *cmd)
{
  int offset = 0;
  char *ptr;
  ALIST *head;

  head = alist;

  // generate offset for shells such as bash etc.
  if ((strstr(alist->alias, "alias ") == alist->alias))
    offset = strlen("alias");

  // walk through the alias list
  while (head && head->alias)
  {
    ptr = head->alias + offset;

    // delete leading blank characters
    while(isblank(*ptr))
      ptr++;

    // check if the alias matches the argument
    if (ptr == strstr(ptr, cmd))
      if ( ! isalpha(*(ptr + strlen(cmd))))
        return ptr;

    head = head->next;
  }

  // argument is not an alias
  return 0;
}

void
free_list(ALIST *alist)
{
  ALIST *head;

  head = alist->next;

  while (alist)
  {
    free(alist);
    alist = head;
    if (head)
      head = head->next;
  }
}

void usage(void)
{
  printf("\nUsage:\tcwhich [-apsv] name ...\n");
  printf("\twhich -h\n");
  printf("\nOptions:");
  printf("\n\t-a\tSearch along the user's \"path\" until all");
  printf("\n\t\tinstances of the executable are found.");
  printf("\n\t-h\tPrint this help message.");
  printf("\n\t-p\tReturns the \"path\" and \"aliases\" defined");
  printf("\n\t\tby a non-interactive shell.");
  printf("\n\t-s\tSuppress messages to standard output.");
  printf("\n\t-v\tRun verbose enabling output of error messages");
  printf("\n\t\tto standard output.");
  printf("\n\tname\tThe name of one or more executables.\n\n");
}

